import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';
import 'package:bmi_global/utils/parsers.dart';

class FirestoreChanges {
  final FirebaseFirestore firestoreInstance;

  FirestoreChanges(this.firestoreInstance);

  Future<FirestoreErrors?> updateDoc({
    required String docPath,
    Map<String, Object?>? data,
    Iterable<String>? deletePaths,
  }) async {
    data ??= {};
    try {
      data.removeWhere((key, value) => value == null);
      if (deletePaths != null) {
        data.addEntries(
          deletePaths.map((e) => MapEntry(e, FieldValue.delete())),
        );
      }
      if (data.isNotEmpty) {
        await firestoreInstance.doc(docPath).update(data);
      }
    } catch (error, stackTrace) {
      return FirestoreErrors(error, stackTrace);
    }
    return null;
  }

  Future<FirestoreErrors?> addDoc({
    required String collPath,
    Map<String, Object?> data = const {},
  }) async {
    try {
      data.removeWhere((key, value) => value == null);
      await firestoreInstance.collection(collPath).add(data);
    } catch (error, stackTrace) {
      return FirestoreErrors(error, stackTrace);
    }
    return null;
  }

  Future<FirestoreErrors?> removeDoc({
    required String docPath,
  }) async {
    try {
      await firestoreInstance.doc(docPath).delete();
    } catch (error, stackTrace) {
      return FirestoreErrors(error, stackTrace);
    }
    return null;
  }
}

class FirestoreErrors {
  Object error;
  StackTrace stackTrace;

  FirestoreErrors(this.error, this.stackTrace);

  void showAlertDialog({
    String? title,
    String? message,
    required BuildContext context,
  }) {
    /**
     * if (title = null && message = null) title = error, message = stackTrace
     * else if (message = null) message = stackTrace
     * else if (title = null) title = error
     * else;
     */
    message ??= title == null ? stackTrace.toString() : error.toString();
    title ??= error.toString();
    showDialog(
      context: context,
      builder: (context) {
        return AlertDialog(
          title: Text(title!),
          content: Text(message!),
          actions: [
            TextButton(
              onPressed: () {
                Navigator.pop(context);
              },
              child: const Text("Ok"),
            ),
          ],
        );
      },
    );
  }
}

abstract class FetchStaticFirebaseDoc<T> {
  T? _doc;
  Object? _error;
  var _loading = true;
  final String path;
  final void Function() notifyListeners;
  final T Function(RawMap) parse;

  bool get loading => _loading;
  T? get doc => _doc;
  Object? get error => _error;

  FetchStaticFirebaseDoc({
    required this.path,
    required this.notifyListeners,
    required this.parse,
  }) {
    _fetchDoc();
  }

  FirebaseFirestore get firestoreInstance;

  void _fetchDoc() async {
    try {
      final doc = await firestoreInstance.doc(path).get();
      final data = doc.data();
      _doc = data == null ? null : parse(RawMap(data));
      _error = null;
    } catch (error) {
      _error = error;
    }
    _loading = false;
    notifyListeners();
  }

  void refresh() {
    if (_loading) return;
    _loading = true;
    notifyListeners();
    _fetchDoc();
  }
}

abstract class StreamFirebaseDoc<T> {
  T? _doc;
  var _loading = true;
  Object? _error;
  final String path;
  final void Function(T? before, T? after) onChange;
  final bool Function(Object error)? cancleOnError;
  final T Function(RawMap data) parse;
  late final void Function() cancle;

  FirebaseFirestore get firestoreInstance;

  bool get loading => _loading;
  T? get doc => _doc;
  Object? get error => _error;

  StreamFirebaseDoc({
    required this.path,
    required this.parse,
    required this.onChange,
    this.cancleOnError,
  }) {
    cancle = firestoreInstance
        .doc(path)
        .snapshots()
        .handleError(_onError)
        .listen(_listen, cancelOnError: false)
        .cancel;
  }

  void _onError(Object error) {
    _loading = false;
    _error = error;
    if (cancleOnError?.call(error) ?? false) cancle();
  }

  T? convertor(Map<String, Object?>? data) {
    if (data == null) return null;
    return parse(RawMap(data));
  }

  void _listen(DocumentSnapshot<Map<String, Object?>> event) {
    _loading = false;
    final data = convertor(event.data());
    onChange(_doc, data);
    _doc = data;
  }
}
